/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.rendering.internal.wiki;

import java.lang.reflect.Type;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.jmock.Expectations;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.xwiki.bridge.DocumentAccessBridge;
import org.xwiki.bridge.SkinAccessBridge;
import org.xwiki.model.EntityType;
import org.xwiki.model.reference.AttachmentReference;
import org.xwiki.model.reference.DocumentReference;
import org.xwiki.model.reference.EntityReferenceResolver;
import org.xwiki.model.reference.EntityReferenceSerializer;
import org.xwiki.rendering.internal.configuration.XWikiRenderingConfiguration;
import org.xwiki.rendering.internal.resolver.DefaultResourceReferenceEntityReferenceResolver;
import org.xwiki.rendering.listener.reference.AttachmentResourceReference;
import org.xwiki.rendering.listener.reference.DocumentResourceReference;
import org.xwiki.rendering.listener.reference.ResourceReference;
import org.xwiki.rendering.listener.reference.ResourceType;
import org.xwiki.rendering.wiki.WikiModel;
import org.xwiki.test.jmock.AbstractMockingComponentTestCase;
import org.xwiki.test.jmock.annotation.MockingRequirement;

/**
 * Unit tests for {@link XWikiWikiModel}.
 * 
 * @version $Id: 3739aec8a2b7e4cfe1619809cc071761b7d7369f $
 * @since 2.0M1
 */
@MockingRequirement(XWikiWikiModel.class)
public class XWikiWikiModelTest extends AbstractMockingComponentTestCase<WikiModel>
{
    private EntityReferenceResolver<ResourceReference> referenceResolver;

    private DocumentAccessBridge documentAccessBridge;

    @Before
    @Override
    public void setUp() throws Exception
    {
        super.setUp();

        this.referenceResolver =
            getComponentManager().getInstance(DefaultResourceReferenceEntityReferenceResolver.TYPE_RESOURCEREFERENCE);
        this.documentAccessBridge = getComponentManager().getInstance((Type) DocumentAccessBridge.class);
    };

    @Test
    public void testGetDocumentEditURLWhenNoQueryStringSpecified() throws Exception
    {
        final DocumentResourceReference reference = new DocumentResourceReference("TargetSpace.TargetPage");
        reference.setAnchor("anchor");

        final EntityReferenceSerializer<?> compactEntityReferenceSerializer =
            getComponentManager().getInstance(EntityReferenceSerializer.TYPE_STRING, "compactwiki");


        // Note: we use a character that needs to be encoded in the current document's page name to make sure the
        // generate query string is encoded.
        final DocumentReference currentDocumentReference = new DocumentReference("Wiki", "Space", "Page\u20AC");

        final DocumentReference documentReference = new DocumentReference("TargetWiki", "TargetSpace", "TargetPage");

        getMockery().checking(new Expectations()
        {
            {
                oneOf(documentAccessBridge).getCurrentDocumentReference();
                will(returnValue(currentDocumentReference));
                oneOf(compactEntityReferenceSerializer).serialize(currentDocumentReference);
                will(returnValue("Wiki:Space.Page\u20AC"));

                oneOf(referenceResolver).resolve(reference, EntityType.DOCUMENT);
                will(returnValue(documentReference));

                // The test is here: we verify that getDocumentURL is called with the query string already encoded
                // since getDocumentURL() doesn't encode it.
                oneOf(documentAccessBridge).getDocumentURL(documentReference, "create",
                    "parent=Wiki%3ASpace.Page%E2%82%AC", "anchor");
            }
        });

        getMockedComponent().getDocumentEditURL(reference);
    }

    /**
     * Tests that the proper image URL is generated when both the width and the height image parameters are specified.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenBothWidthAndHeightAttributesAreSpecified() throws Exception
    {
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("width", "100px");
        parameters.put("height", "50");

        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true, "width=100&height=50");
    }

    /**
     * Tests that the proper image URL is generated when both the width and the height image parameters are specified
     * but including them in the image URL is forbidden from the rendering configuration.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenIncludingImageDimensionsIsForbidden() throws Exception
    {
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("width", "101px");
        parameters.put("height", "55px");

        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, false, null);
    }

    /**
     * Tests that the proper image URL is generated when both the width and the height CSS properties are specified.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenBothWidthAndHeightCSSPropertiesAreSpecified() throws Exception
    {
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("style", "border: 1px; height: 30px; margin-top: 2em; width: 70px");

        AttachmentResourceReference reference = new AttachmentResourceReference("attachmentReference");
        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true, "width=70&height=30");
    }

    /**
     * Tests that the proper image URL is generated when only the width image parameter is specified.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenOnlyWidthAttributeIsSpecified() throws Exception
    {
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("width", "150");
        parameters.put("height", "30%");

        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true, "width=150");
    }

    /**
     * Tests that the proper image URL is generated when only the height CSS property is specified.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenOnlyHeightCSSPropertyIsSpecified() throws Exception
    {
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("style", "width: 5cm; height: 80px");

        AttachmentResourceReference reference = new AttachmentResourceReference("attachmentReference");
        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true, "height=80");
    }

    /**
     * Tests that the proper image URL is generated when both the width and the height are unspecified and image size is
     * limited in the configuration.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenBothWidthAndHeightAreUnspecifiedAndImageSizeIsLimited() throws Exception
    {
        final XWikiRenderingConfiguration configuration =
            getComponentManager().getInstance((Type) XWikiRenderingConfiguration.class);
        getMockery().checking(new Expectations()
        {
            {
                oneOf(configuration).getImageWidthLimit();
                will(returnValue(200));
                oneOf(configuration).getImageHeightLimit();
                will(returnValue(170));
            }
        });
        Map<String, String> parameters = Collections.emptyMap();

        AttachmentResourceReference reference = new AttachmentResourceReference("attachmentReference");
        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true,
            "width=200&height=170&keepAspectRatio=true");
    }

    /**
     * Tests that the proper image URL is generated when both the width and the height are unspecified and only the
     * image width is limited in the configuration.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenBothWidthAndHeightAreUnspecifiedAndOnlyImageWidthIsLimited() throws Exception
    {
        final XWikiRenderingConfiguration configuration =
            getComponentManager().getInstance((Type) XWikiRenderingConfiguration.class);
        getMockery().checking(new Expectations()
        {
            {
                oneOf(configuration).getImageWidthLimit();
                will(returnValue(25));
                oneOf(configuration).getImageHeightLimit();
                will(returnValue(-1));
            }
        });
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("width", "45%");
        parameters.put("style", "height:10em");

        AttachmentResourceReference reference = new AttachmentResourceReference("attachmentReference");
        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true, "width=25");
    }

    /**
     * Tests that the proper image URL is generated when both the width and the height are unspecified and the image
     * size is not limited in the configuration.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenBothWidthAndHeightAreUnspecifiedAndImageSizeIsNotLimited() throws Exception
    {
        final XWikiRenderingConfiguration configuration =
            getComponentManager().getInstance((Type) XWikiRenderingConfiguration.class);
        getMockery().checking(new Expectations()
        {
            {
                oneOf(configuration).getImageWidthLimit();
                will(returnValue(-1));
                oneOf(configuration).getImageHeightLimit();
                will(returnValue(-1));
            }
        });
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("style", "bad CSS declaration");

        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true, null);
    }

    /**
     * Tests that the proper image URL is generated when the attachment URL has a query string and a fragment
     * identifier.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenImageReferenceHasQueryString() throws Exception
    {
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("height", "17");

        AttachmentResourceReference reference = new AttachmentResourceReference("attachmentReference");
        reference.setParameter(AttachmentResourceReference.QUERY_STRING, "param=value");
        testImageURL(reference, parameters, true, "param=value&height=17");
    }

    /**
     * Tests that the proper image URL is generated when both the style and the dimension parameters are specified.
     * 
     * @throws Exception if an exception occurs while running the test
     */
    @Test
    public void testGetImageURLWhenBothStyleAndDimensionParametersAreSpecified() throws Exception
    {
        Map<String, String> parameters = new HashMap<String, String>();
        parameters.put("height", "46");
        parameters.put("width", "101px");
        parameters.put("style", "width: 20%; height:75px");

        // Note that the style parameter take precedence over the dimension parameters and the width is actually 20% but
        // we can't use it for resizing the image on the server side so it's omitted from the query string.
        testImageURL(new AttachmentResourceReference("attachmentReference"), parameters, true, "height=75");
    }

    @Test
    public void testGetImageURLWhenIcon() throws Exception
    {
        ResourceReference reference = new ResourceReference("iconname", ResourceType.ICON);
        final SkinAccessBridge skinAccessBridge = getComponentManager().getInstance((Type) SkinAccessBridge.class);
        getMockery().checking(new Expectations()
        {
            {
                oneOf(skinAccessBridge).getIconURL("iconname");
                will(returnValue("/path/to/icon"));
            }
        });

        Assert.assertEquals("/path/to/icon",
            getMockedComponent().getImageURL(reference, Collections.<String, String>emptyMap()));
    }

    @Test
    public void testGetDocumentViewURLWhenNoBaseReference() throws Exception
    {
        final ResourceReference reference = new ResourceReference("reference", ResourceType.DOCUMENT);

        final DocumentReference documentReference = new DocumentReference("wiki", "space", "page");
        getMockery().checking(new Expectations()
        {
            {
                oneOf(referenceResolver).resolve(reference, EntityType.DOCUMENT);
                will(returnValue(documentReference));
                oneOf(documentAccessBridge).getDocumentURL(documentReference, "view", null, null);
                will(returnValue("viewurl"));
            }
        });

        Assert.assertEquals("viewurl", getMockedComponent().getDocumentViewURL(reference));
    }

    @Test
    public void testGetDocumentViewURLWhenBaseReferenceSpecified() throws Exception
    {
        final ResourceReference reference = new ResourceReference("reference", ResourceType.DOCUMENT);
        reference.addBaseReference("base");

        final DocumentReference documentReference = new DocumentReference("wiki", "space", "page");
        getMockery().checking(new Expectations()
        {
            {
                oneOf(referenceResolver).resolve(reference, EntityType.DOCUMENT);
                will(returnValue(documentReference));
                oneOf(documentAccessBridge).getDocumentURL(documentReference, "view", null, null);
                will(returnValue("viewurl"));
            }
        });

        Assert.assertEquals("viewurl", getMockedComponent().getDocumentViewURL(reference));
    }

    private void testImageURL(final ResourceReference imageReference, Map<String, String> parameters,
        final boolean expectedIsImageDimensionsIncludedInImageURL, final String expectedQueryString) throws Exception
    {
        final AttachmentReference attachmentReference =
            new AttachmentReference("image", new DocumentReference("wiki", "space", "page"));
        final XWikiRenderingConfiguration configuration =
            getComponentManager().getInstance((Type) XWikiRenderingConfiguration.class);
        getMockery().checking(new Expectations()
        {
            {
                oneOf(configuration).isImageDimensionsIncludedInImageURL();
                will(returnValue(expectedIsImageDimensionsIncludedInImageURL));
                oneOf(referenceResolver).resolve(imageReference, EntityType.ATTACHMENT);
                will(returnValue(attachmentReference));
                oneOf(documentAccessBridge).getAttachmentURL(attachmentReference, expectedQueryString, false);
                will(returnValue("?" + expectedQueryString));
            }
        });

        Assert.assertEquals("?" + expectedQueryString, getMockedComponent().getImageURL(imageReference, parameters));
    }
}
